require "defines"

-- always on the lookout for trains, even when moving
COWARD_MODE = false
-- enough self-control to keep quiet
QUIET_MODE = false

-- how far each side of the player to look for track
TRACK_RANGE = 0.75
-- minimum time in ticks between yelps
YELP_COOLDOWN = 300
-- how far each side of player to look for trains
SEARCH_RANGE = 4
-- angle from train path within which to take evasive action
SEARCH_ANGLE = 1/8

-- plays sound at the location of the given entity
-- sounds are played using a dedicated explosion-type entity for each sound file
function playSoundAtEntity(sound, entity)
    entity.surface.create_entity({name = sound, position = entity.position})
end

-- returns bearing from e to t as 0.0 to 1.0, where 0 is north
function getBearing(e_x, e_y, t_x, t_y)
    return (math.atan2(e_x - t_x, t_y - e_y) / math.pi + 1) / 2
end

-- returns signed difference of b2 from b1, as -0.5 to 0.5
function getRelativeBearing(b1, b2)
    return ((b2 - b1 + 0.5) % 1) - 0.5
end

-- converts vehicle orientation to player/entity rotation, rounded to nearest
function orientation2direction(o)
    return math.floor(o * 8 + 0.5) % 8
end

-- counter for yelp cooldown
yelp = {}

script.on_event(defines.events.on_tick, function()
    local thisPlayer = nil
    local players = game.players
    for i=1, #players do
        thisPlayer = players[i]
        if (thisPlayer.connected) then
            if (thisPlayer.character) then
                if not yelp[i] then
                    yelp[i] = 0
                end
                if yelp[i] > 0 then
                    yelp[i] = yelp[i] - 1
                end
                
                if (COWARD_MODE or not thisPlayer.walking_state.walking) and not thisPlayer.vehicle then
                    local trains = thisPlayer.surface.find_entities_filtered
                    {
                        area =
                            {{thisPlayer.position.x-SEARCH_RANGE, thisPlayer.position.y-SEARCH_RANGE},
                            {thisPlayer.position.x+SEARCH_RANGE, thisPlayer.position.y+SEARCH_RANGE}},
                        type= "locomotive"
                    }
                    if #trains > 0 then
                        local train = trains[1].train
                        if train.state == 0 or train.state == 4 or train.state == 6 then
                            local straight = thisPlayer.surface.find_entities_filtered
                            {
                                area =
                                    {{thisPlayer.position.x-TRACK_RANGE, thisPlayer.position.y-TRACK_RANGE},
                                    {thisPlayer.position.x+TRACK_RANGE, thisPlayer.position.y+TRACK_RANGE}},
                                type= "straight-rail"
                            }
                            local curved = thisPlayer.surface.find_entities_filtered
                            {
                                area =
                                    {{thisPlayer.position.x-TRACK_RANGE, thisPlayer.position.y-TRACK_RANGE},
                                    {thisPlayer.position.x+TRACK_RANGE, thisPlayer.position.y+TRACK_RANGE}},
                                type= "curved-rail"
                            }
                            if #straight > 0 or #curved > 0 then
                                local loco = trains[1]
                                local bearing = getBearing(loco.position.x, loco.position.y, thisPlayer.position.x, thisPlayer.position.y)
                                local deviance = getRelativeBearing(bearing, loco.orientation)
                                if math.abs(deviance) < SEARCH_ANGLE then
                                    local run_direction
                                    if deviance < 0 then
                                        run_direction = orientation2direction(loco.orientation + 0.25)
                                    else
                                        run_direction = orientation2direction(loco.orientation - 0.25)
                                    end
                                    thisPlayer.walking_state = {walking = true, direction = run_direction}
                                    if not QUIET_MODE and yelp[i] == 0 then
                                        playSoundAtEntity("scared", thisPlayer)
                                        yelp[i] = YELP_COOLDOWN
                                    end
                                end
                            end
                        end
                    end
                end
            end
        end
    end
end)